require 'socket'
require 'uri'
require 'json'
require 'socket'

module Rack
  class LogstashWriter
    DEFAULT_OPTIONS = {
        request_headers: nil,
        response_headers: nil,
        statuses: [*(500..600)],
        body_len: 1000,
        add_curl_command?: false,
        add_request_script?: false,
        curl_opts: {}
    }
    REQUIRED_OPTIONS = [:url]
    CURL_DEFAULT_OPTS = {
        data_key: 'rack.request.form_hash',
        data_proc: proc(&:to_json),
        ignored_headers: [],
        verbose: true
    }
    COMMON_ENCODINGS = %w(compress deflate exi gzip pack200-gzip br)

    # Initialize a new Rack adapter, logstash writer
    # @param [Proc] app
    # @param [Hash] options, defaults can be viewed at ::DEFAULT_OPTIONS
    # @option options [String] :url: required, udp and files schemes are also available. no default values.
    # @option options [Hash] :request_headers: optional, parameters to add to the report from the request headers. default nil
    # @option options [Hash] :response_headers: optional, parameters to add to the report from the responce headers. default nil
    # @option options [Fixnum] :body_len: optional, include the first given chars from the body. default 1000
    # @option options [Array] :statuses: optional, send events to log stash only for those statuses. default [*(500..600)]
    # @option options [Proc] :proc: optional, the function will call the proc with the env and call
    # @option options [Boolean] :add_curl_command?: add [:curl] to the log with *curl* unix command string set to mock the request that led to the log
    # @option options [Boolean] :add_request_script?: alias for the :curl: @option
    def initialize app, options = {}
      validate_options options

      @app = app
      @options = DEFAULT_OPTIONS.merge(options)
      @options[:url]= URI(@options[:url])
      @proc = @options[:proc]
    end

    # Call to the app cal and log if the returned status is in the array of return data.
    # @param [Hash] env : the environment
    # @return [Array] [status:Fixnum, headers:Hash, body:Rack::BodyProxy]
    def call env
      began = Time.now
      s, h, b = @app.call(env)
      b = BodyProxy.new(b) { log(env, s, h, began, b) } if @options[:statuses].include?(s.to_i)
      [s, h, b]
    end

    # Return the correct connection by the uri - udp/tcp/file
    private

    def validate_options(options)
      missing_required_options = REQUIRED_OPTIONS - options.keys
      raise "#{missing_required_options} options must be given!" unless missing_required_options.empty?
    end

    # @return [File|UDPSocket|TCPSocket] new instance of socket of file for data writing/sending.
    # raising an error if url-scheme could not be identified
    def device
      @device ||=
          case @options[:url].scheme
            when 'file' then
              ::File.new(@options[:url].path, "a").tap { |f| f.sync = true }
            when 'udp' then
              UDPSocket.new.tap { |s| s.connect @options[:url].host, @options[:url].port }
            when 'tcp' then
              TCPSocket.new @options[:url].host, @options[:url].port
            else
              raise "Unknown scheme #{@options[:url].scheme}"
          end
    end

    # @param [Hash] env as given from Rack
    # @param [Fixnum] status
    # @param [Hash] response_headers
    # @param [Time|DateTime|Date] began_at
    # @param [Object] body of the request
    # Log the data to the device
    def log env, status, response_headers, began_at, body
      data = {
          method: env['REQUEST_METHOD'],
          path: env['PATH_INFO'],
          query_string: env['QUERY_STRING'],
          host: Socket.gethostname,
          status: status.to_i,
          duration: (Time.now - began_at),
          remote_addr: env['REMOTE_ADDR'],
          request: request_line(env),
          :'X-Forwarded-For' => response_headers['X-Forwarded-For']
      }

      # Added calling for the proc and merge the data if it exists
      if @proc
        begin
          new_hash = @proc.call(env)
          data = data.merge new_hash if new_hash.is_a?(Hash)
        rescue Exception => e
          STDERR.puts "Exception in your proc : #{e.message}."
        end
      end

      # add curl if users enables it
      if @options[:add_curl_command?] || @options[:add_request_script?]
        data[:curl] = gen_curl_by_rack_env(env, response_headers, CURL_DEFAULT_OPTS.merge(@options[:curl_opts]))
      end

      # This just works for all body types (magic?)... see http://www.rubydoc.info/github/rack/rack/Rack/BodyProxy
      body.each { |x| data[:body] = x[0..@options[:body_len]] }

      unless @options[:request_headers].nil?
        @options[:request_headers].each do |header, log_key|
          env_key = "HTTP_#{header.upcase.gsub('-', '_')}"
          data[log_key] = env[env_key] if env[env_key]
        end
      end
      unless @options[:response_headers].nil?
        @options[:response_headers].each do |header, log_key|
          data[log_key] = response_headers[header] if response_headers[header]
        end
      end

      data[:error_msg] = env['sinatra.error'] if env.has_key?('sinatra.error')

      unless @options[:body_regex].nil?
        @options[:body_regex].each do |k, v|
          data[k] = data[:body].to_s.match(/#{v}/).captures[0].gsub('\\', '').gsub('"', '') rescue data[k] = ''
        end
      end

      severity = case status
                   when 300..399
                     'WARN'
                   when 400..599
                     'ERROR'
                   else
                     'DEBUG'
                 end
      data = {severity: severity}.merge(data)

      begin
        device.puts data.to_json
      rescue Exception => e
        STDERR.puts "Error : Failed to write log to : #{@options[:url]}, #{e.message}."
        @device = nil
      end
    end

    # @param [Hash] env rack env obj
    # @return [String] e.g. "METHOD name/path/path?query=query protocol"
    def request_line env
      line = "#{env['REQUEST_METHOD']} #{env['SCRIPT_NAME']}#{env['PATH_INFO']}"
      line << "?#{env['QUERY_STRING']}" if env['QUERY_STRING'] && !env['QUERY_STRING'].empty?
      line << " #{env['SERVER_PROTOCOL']}"
    end

    # @param [Hash] env as received from Rack
    # @param [Hash] opts
    # * @option [Proc] :data_proc: for processing the data prior for inserting it to query (by --data flag)
    # * @option [Array<String>] :ignored_headers: to prevent inclusion to the query (by --header flag)
    # * @option [Array<String>] :extra_arguments: string ot args to add to the query
    # * @option [Boolean] :verbose: add --verbose switch if true. default: true
    # * @option [Boolean] :compressed: add or exclude --compressed switch by force. default: chooses automatically.
    # @return [String] of curl unix-command that will imitate the request to server/service
    def gen_curl_by_rack_env env, response_headers, opts
      command_acc = ["curl #{url(env)} "]

      command_acc << "--request #{env['REQUEST_METHOD']}"

      # form-data will be processed by the proc sent by the user
      data = opts[:data_proc].call(env[opts[:data_key]]) rescue env['rack.request.form_hash']
      command_acc << "--data #{data.inspect}" if data && !data.empty?

      # add all headers to the query. do not add +ignore_headers+ specified by the user

      env.keys.select { |env_key| env_key.to_s.start_with?('HTTP_') || env_key.to_s.start_with?('CONTENT_') }
          .each do |env_key|
        header_name = env_key.to_s.sub(/^HTTP_/i, '').split('_').map(&:capitalize).join('-')
        next if opts[:ignored_headers].include?(header_name)
        header_value = env[env_key].inspect[1...-1]
        command_acc << "--header \"#{header_name}: #{header_value}\""
      end

      # the curl script is intended for debugging, therefore --verbose switch is added (unless specified by caller)
      command_acc << '--verbose' if opts[:verbose]
      encodings = response_headers['Content-Encoding'].to_s.split(',').map(&:strip) & COMMON_ENCODINGS
      command_acc << '--compressed' if (opts[:compressed].nil? ? !encodings.empty? : opts[:compressed])

      # let user change the query as they desire
      command_acc += opts[:extra_arguments] if opts[:extra_arguments]


      command_acc.join " \\\n"
    end

    # @param [Hash] env rack env
    # @return [String] url-scheme://prefix.name.suffix:port/uri/uri/uri?query=query:query=queryrevert master
    def url env
      url = "#{env['rack.url_scheme']}://#{env['SERVER_NAME']}:#{env['SERVER_PORT']}#{env['REQUEST_URI']}"
      query_string = env['QUERY_STRING'].to_s.strip
      !query_string.empty? ? url << "?#{query_string}" : url
    end
  end
end
